# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Tests for the CLI commands."""

from __future__ import annotations

import csv
import io
import json
import textwrap

import pytest

from provero import __version__
from provero.cli import main as cli_main
from provero.cli.main import _print_csv, app
from provero.core.results import CheckResult, Severity, Status, SuiteResult


class TestVersion:
    def test_shows_version(self, cli_runner):
        result = cli_runner.invoke(app, ["version"])
        assert result.exit_code == 0
        assert __version__ in result.output


class TestInit:
    def test_creates_template(self, cli_runner, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = cli_runner.invoke(app, ["init"])
        assert result.exit_code == 0
        content = (tmp_path / "provero.yaml").read_text()
        assert "source:" in content

    def test_existing_file_aborts(self, cli_runner, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        (tmp_path / "provero.yaml").write_text("existing")
        result = cli_runner.invoke(app, ["init"])
        assert result.exit_code == 1
        assert "already exists" in result.output

    def test_from_source_bad_format(self, cli_runner, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = cli_runner.invoke(app, ["init", "--from-source", "bad_format"])
        assert result.exit_code == 1
        assert "Format" in result.output


class TestRun:
    def test_missing_config(self, cli_runner, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = cli_runner.invoke(app, ["run", "--config", "nonexistent.yaml"])
        assert result.exit_code == 1
        assert "not found" in result.output

    def test_passing_checks(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(app, ["run", "--config", config_path, "--no-store"])
        assert result.exit_code == 0

    def test_failing_checks(self, cli_runner, duckdb_file, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        config = tmp_path / "fail.yaml"
        config.write_text(
            textwrap.dedent(f"""\
            source:
              type: duckdb
              connection: "{duckdb_file}"
              table: orders

            checks:
              - unique: customer_id
        """)
        )
        result = cli_runner.invoke(app, ["run", "--config", str(config), "--no-store"])
        assert result.exit_code == 1

    def test_json_format(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app, ["run", "--config", config_path, "--format", "json", "--no-store"]
        )
        assert result.exit_code == 0
        # Output should contain valid JSON (may have Rich markup too, parse the JSON part)
        # The JSON is printed via console.print, find the JSON object in output
        output = result.output
        start = output.index("{")
        end = output.rindex("}") + 1
        parsed = json.loads(output[start:end])
        assert "suite_name" in parsed

    def test_csv_format(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app, ["run", "--config", config_path, "--format", "csv", "--no-store"]
        )
        assert result.exit_code == 0

        rows = list(csv.DictReader(io.StringIO(result.output)))
        assert len(rows) > 0
        assert list(rows[0].keys()) == [
            "suite_name",
            "check_type",
            "column",
            "status",
            "severity",
            "observed_value",
            "expected_value",
        ]
        assert rows[0]["suite_name"]
        assert rows[0]["check_type"]
        assert rows[0]["status"]
        assert rows[0]["severity"]

    def test_csv_format_none_values_render_empty(self, capsys):
        result = SuiteResult(
            suite_name="suite_a",
            status=Status.PASS,
            checks=[
                CheckResult(
                    check_name="custom_sql_check",
                    check_type="custom_sql",
                    status=Status.PASS,
                    severity=Severity.CRITICAL,
                    column=None,
                    observed_value=None,
                    expected_value=None,
                )
            ],
        )

        _print_csv(result, include_header=True)
        output = capsys.readouterr().out
        rows = list(csv.DictReader(io.StringIO(output)))

        assert len(rows) == 1
        assert rows[0]["suite_name"] == "suite_a"
        assert rows[0]["column"] == ""
        assert rows[0]["observed_value"] == ""
        assert rows[0]["expected_value"] == ""

    def test_csv_format_multi_suite_writes_header_once(
        self, cli_runner, duckdb_file, tmp_path, monkeypatch
    ):
        monkeypatch.chdir(tmp_path)
        config = tmp_path / "multi_suite.yaml"
        config.write_text(
            textwrap.dedent(f"""\
            version: "1.0"
            sources:
              warehouse:
                type: duckdb
                connection: "{duckdb_file}"

            suites:
              - name: orders_suite
                source: warehouse
                table: orders
                checks:
                  - not_null: order_id

              - name: events_suite
                source: warehouse
                table: events
                checks:
                  - not_null: event_id
        """)
        )

        result = cli_runner.invoke(
            app, ["run", "--config", str(config), "--format", "csv", "--no-store"]
        )
        assert result.exit_code == 0

        header = "suite_name,check_type,column,status,severity,observed_value,expected_value"
        assert result.output.count(header) == 1

        rows = list(csv.DictReader(io.StringIO(result.output)))
        assert len(rows) >= 2
        suite_names = {row["suite_name"] for row in rows}
        assert "orders_suite" in suite_names
        assert "events_suite" in suite_names

    def test_html_report(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app, ["run", "--config", config_path, "--report", "html", "--no-store"]
        )
        assert result.exit_code == 0
        # Check that HTML report was generated somewhere under .provero/reports/
        reports = list(tmp_path.glob(".provero/reports/*.html"))
        assert len(reports) >= 1

    def test_no_store(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(app, ["run", "--config", config_path, "--no-store"])
        assert result.exit_code == 0
        assert not (tmp_path / ".provero" / "results.db").exists()

    def test_no_optimize(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app, ["run", "--config", config_path, "--no-optimize", "--no-store"]
        )
        assert result.exit_code == 0


class TestValidate:
    def test_valid_config(self, cli_runner, sample_config_file):
        result = cli_runner.invoke(app, ["validate", "--config", str(sample_config_file)])
        assert result.exit_code == 0
        assert "Valid" in result.output

    def test_invalid_config(self, cli_runner, tmp_path):
        bad_config = tmp_path / "bad.yaml"
        bad_config.write_text("not_a_valid_key: true\n")
        result = cli_runner.invoke(app, ["validate", "--config", str(bad_config)])
        assert result.exit_code == 1


class TestContract:
    def test_contract_validate(self, cli_runner, duckdb_file, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        config = tmp_path / "contract.yaml"
        config.write_text(
            textwrap.dedent(f"""\
            version: "1.0"
            sources:
              warehouse:
                type: duckdb
                connection: "{duckdb_file}"

            suites:
              - name: orders_checks
                source: warehouse
                table: orders
                checks:
                  - not_null: order_id

            contracts:
              - name: orders_contract
                source: warehouse
                table: orders
                schema:
                  columns:
                    - name: order_id
                      type: integer
                    - name: customer_id
                      type: varchar
                    - name: amount
                      type: decimal
                    - name: status
                      type: varchar
        """)
        )
        result = cli_runner.invoke(app, ["contract", "validate", "--config", str(config)])
        assert result.exit_code == 0

    def test_contract_diff(self, cli_runner, tmp_path):
        old_config = tmp_path / "old.yaml"
        old_config.write_text(
            textwrap.dedent("""\
            source:
              type: duckdb
              table: orders

            checks:
              - not_null: order_id

            contracts:
              - name: orders_contract
                table: orders
                schema:
                  columns:
                    - name: order_id
                      type: integer
        """)
        )
        new_config = tmp_path / "new.yaml"
        new_config.write_text(
            textwrap.dedent("""\
            source:
              type: duckdb
              table: orders

            checks:
              - not_null: order_id

            contracts:
              - name: orders_contract
                table: orders
                schema:
                  columns:
                    - name: order_id
                      type: integer
                    - name: name
                      type: varchar
        """)
        )
        result = cli_runner.invoke(app, ["contract", "diff", str(old_config), str(new_config)])
        assert result.exit_code == 0
        assert "orders_contract" in result.output


class TestQuietFlag:
    """Tests for the --quiet / -q flag."""

    @pytest.fixture(autouse=True)
    def _reset_quiet_flag(self):
        yield
        cli_main._quiet = False

    def test_quiet_run_table_suppresses_output(
        self, cli_runner, duckdb_config_file, monkeypatch, tmp_path
    ):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(app, ["--quiet", "run", "--config", config_path, "--no-store"])
        assert result.exit_code == 0
        # Table output (Suite:, Score:, PASS) must be absent
        assert "Suite:" not in result.output
        assert "Score:" not in result.output

    def test_quiet_short_flag(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(app, ["-q", "run", "--config", config_path, "--no-store"])
        assert result.exit_code == 0
        assert "Suite:" not in result.output

    def test_quiet_json_still_emits_json(
        self, cli_runner, duckdb_config_file, monkeypatch, tmp_path
    ):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app,
            [
                "--quiet",
                "run",
                "--config",
                config_path,
                "--format",
                "json",
                "--no-store",
            ],
        )
        assert result.exit_code == 0
        # Should still contain valid JSON
        output = result.output.strip()
        parsed = json.loads(output)
        assert "suite_name" in parsed

    def test_quiet_csv_still_emits_csv(self, cli_runner, duckdb_config_file, monkeypatch, tmp_path):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(
            app,
            [
                "--quiet",
                "run",
                "--config",
                config_path,
                "--format",
                "csv",
                "--no-store",
            ],
        )
        assert result.exit_code == 0
        rows = list(csv.DictReader(io.StringIO(result.output)))
        assert len(rows) > 0
        assert "suite_name" in rows[0]

    def test_quiet_validate_suppresses_info(self, cli_runner, sample_config_file):
        result = cli_runner.invoke(
            app, ["--quiet", "validate", "--config", str(sample_config_file)]
        )
        assert result.exit_code == 0
        # "Valid." and "Schema validation passed" are informational
        assert "Valid." not in result.output
        assert "Schema validation passed" not in result.output

    def test_quiet_init_suppresses_confirmation(self, cli_runner, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = cli_runner.invoke(app, ["--quiet", "init"])
        assert result.exit_code == 0
        assert (tmp_path / "provero.yaml").exists()
        assert "Created" not in result.output

    def test_quiet_preserves_exit_code_on_failure(
        self, cli_runner, duckdb_file, tmp_path, monkeypatch
    ):
        monkeypatch.chdir(tmp_path)
        config = tmp_path / "fail.yaml"
        config.write_text(
            textwrap.dedent(f"""\
            source:
              type: duckdb
              connection: "{duckdb_file}"
              table: orders

            checks:
              - unique: customer_id
        """)
        )
        result = cli_runner.invoke(app, ["--quiet", "run", "--config", str(config), "--no-store"])
        assert result.exit_code == 1

    def test_quiet_profile_suppresses_table(
        self, cli_runner, duckdb_config_file, monkeypatch, tmp_path
    ):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        result = cli_runner.invoke(app, ["--quiet", "profile", "--config", config_path])
        assert result.exit_code == 0
        assert "Profile:" not in result.output
        assert "Column" not in result.output

    def test_quiet_history_suppresses_table(
        self, cli_runner, duckdb_config_file, monkeypatch, tmp_path
    ):
        monkeypatch.chdir(tmp_path)
        config_path = str(duckdb_config_file["config_path"])
        # First run to populate history
        cli_runner.invoke(app, ["run", "--config", config_path])
        # Now check quiet history
        result = cli_runner.invoke(app, ["--quiet", "history"])
        assert result.exit_code == 0
        assert "Run History" not in result.output



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""PostgreSQL connector via SQLAlchemy."""

from __future__ import annotations

from typing import Any, cast

from sqlalchemy import create_engine, text
from sqlalchemy.engine import Engine


class SQLAlchemyConnection:
    """SQLAlchemy-based connection wrapper."""

    def __init__(self, engine: Engine) -> None:
        self._engine = engine
        self._conn = engine.connect()

    def execute(self, query: str, params: dict[str, Any] | None = None) -> list[dict[str, Any]]:
        result = self._conn.execute(text(query), params or {})
        columns = list(result.keys())
        return [dict(zip(columns, row, strict=True)) for row in result.fetchall()]

    def get_columns(self, table: str) -> list[dict[str, Any]]:
        if "." in table:
            schema_name, table_name = table.rsplit(".", 1)
            result = self._conn.execute(
                text(
                    "SELECT column_name, data_type, is_nullable "
                    "FROM information_schema.columns "
                    "WHERE table_schema = :schema AND table_name = :table "
                    "ORDER BY ordinal_position"
                ),
                {"schema": schema_name, "table": table_name},
            )
        else:
            result = self._conn.execute(
                text(
                    "SELECT column_name, data_type, is_nullable "
                    "FROM information_schema.columns "
                    "WHERE table_name = :table "
                    "AND table_schema NOT IN ('information_schema', 'pg_catalog') "
                    "ORDER BY ordinal_position"
                ),
                {"table": table},
            )
        return [
            {"name": row[0], "type": row[1], "nullable": row[2] == "YES"}
            for row in result.fetchall()
        ]

    def close(self) -> None:
        self._conn.close()
        self._engine.dispose()


class PostgresConnector:
    """Connector for PostgreSQL databases."""

    def __init__(self, connection_string: str) -> None:
        self.connection_string = connection_string

    def connect(self) -> SQLAlchemyConnection:
        engine = create_engine(self.connection_string)
        return SQLAlchemyConnection(engine)

    def disconnect(self, connection: SQLAlchemyConnection) -> None:
        connection.close()

    def get_schema(self, connection: SQLAlchemyConnection, table: str) -> list[dict[str, Any]]:
        return connection.get_columns(table)

    def get_profile(
        self,
        connection: SQLAlchemyConnection,
        table: str,
        columns: list[str] | None = None,
        sample_size: int | None = None,
    ) -> dict[str, Any]:
        from provero.core.profiler import profile_table

        result = profile_table(connection, table, sample_size=sample_size)
        data = {
            "table": result.table,
            "row_count": result.row_count,
            "column_count": result.column_count,
            "columns": [
                {
                    "name": c.name,
                    "dtype": c.dtype,
                    "null_count": c.null_count,
                    "null_pct": c.null_pct,
                    "distinct_count": c.distinct_count,
                    "distinct_pct": c.distinct_pct,
                }
                for c in result.columns
            ],
        }
        if columns:
            cols_list = cast(list[dict[str, Any]], data["columns"])
            data["columns"] = [c for c in cols_list if c["name"] in columns]
        return data


class SQLAlchemyConnector:
    """Generic connector for any SQLAlchemy-supported database."""

    def __init__(self, connection_string: str) -> None:
        self.connection_string = connection_string

    def connect(self) -> SQLAlchemyConnection:
        engine = create_engine(self.connection_string)
        return SQLAlchemyConnection(engine)

    def disconnect(self, connection: SQLAlchemyConnection) -> None:
        connection.close()

    def get_schema(self, connection: SQLAlchemyConnection, table: str) -> list[dict[str, Any]]:
        return connection.get_columns(table)

    def get_profile(
        self,
        connection: SQLAlchemyConnection,
        table: str,
        columns: list[str] | None = None,
        sample_size: int | None = None,
    ) -> dict[str, Any]:
        from provero.core.profiler import profile_table

        result = profile_table(connection, table, sample_size=sample_size)
        data = {
            "table": result.table,
            "row_count": result.row_count,
            "column_count": result.column_count,
            "columns": [
                {
                    "name": c.name,
                    "dtype": c.dtype,
                    "null_count": c.null_count,
                    "null_pct": c.null_pct,
                    "distinct_count": c.distinct_count,
                    "distinct_pct": c.distinct_pct,
                }
                for c in result.columns
            ],
        }
        if columns:
            cols_list = cast(list[dict[str, Any]], data["columns"])
            data["columns"] = [c for c in cols_list if c["name"] in columns]
        return data



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""HTML report generator."""

from __future__ import annotations

import importlib.resources
from pathlib import Path
from typing import TYPE_CHECKING

from jinja2 import Environment

from provero import __version__
from provero.core.results import Status

if TYPE_CHECKING:
    from provero.contracts.models import ContractResult
    from provero.core.results import SuiteResult


def generate_html_report(
    suite_result: SuiteResult,
    contract_results: list[ContractResult] | None = None,
    output_path: Path | None = None,
) -> str:
    """Generate a self-contained HTML report from suite results.

    Args:
        suite_result: The suite execution result.
        contract_results: Optional contract validation results.
        output_path: If provided, writes HTML to this file path.

    Returns:
        The rendered HTML string.
    """
    template_text = _load_template()
    env = Environment(autoescape=True)
    template = env.from_string(template_text)

    failing_checks = [
        c
        for c in suite_result.checks
        if c.status in (Status.FAIL, Status.WARN)
        and (c.failing_rows_query or c.failing_rows_sample)
    ]

    checks_data = []
    for c in suite_result.checks:
        checks_data.append(
            {
                "check_type": c.check_type,
                "check_name": c.check_name,
                "column": c.column,
                "status": c.status.value,
                "observed_value": str(c.observed_value) if c.observed_value is not None else "",
                "expected_value": str(c.expected_value) if c.expected_value is not None else "",
                "severity": c.severity.value,
                "failing_rows_query": c.failing_rows_query,
                "failing_rows_sample": c.failing_rows_sample,
            }
        )

    failing_data = []
    for c in failing_checks:
        failing_data.append(
            {
                "check_name": c.check_name,
                "failing_rows_query": c.failing_rows_query,
                "failing_rows_sample": c.failing_rows_sample,
            }
        )

    contract_data = None
    if contract_results:
        contract_data = []
        for cr in contract_results:
            contract_data.append(
                {
                    "contract_name": cr.contract_name,
                    "status": cr.status,
                    "violations": [
                        {"rule": v.rule, "message": v.message, "severity": v.severity}
                        for v in cr.violations
                    ],
                    "schema_drift": [
                        {
                            "column": d.column,
                            "change_type": d.change_type,
                            "expected": d.expected,
                            "actual": d.actual,
                        }
                        for d in cr.schema_drift
                    ],
                }
            )

    html = template.render(
        suite_name=suite_result.suite_name,
        status=suite_result.status.value,
        timestamp=suite_result.started_at.strftime("%Y-%m-%d %H:%M:%S UTC"),
        total=suite_result.total,
        passed=suite_result.passed,
        failed=suite_result.failed,
        warned=suite_result.warned,
        errored=suite_result.errored,
        quality_score=suite_result.quality_score,
        checks=checks_data,
        failing_checks=failing_data,
        contract_results=contract_data,
        version=__version__,
    )

    if output_path:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(html, encoding="utf-8")

    return html


def _load_template() -> str:
    """Load the HTML template from package resources."""
    try:
        ref = importlib.resources.files("provero.reporting").joinpath("templates/report.html")
        return ref.read_text(encoding="utf-8")
    except (FileNotFoundError, TypeError):
        template_path = Path(__file__).parent / "templates" / "report.html"
        return template_path.read_text(encoding="utf-8")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Freshness checks: freshness, latency."""

from __future__ import annotations

import re

from provero.checks.registry import register_check
from provero.connectors.base import Connection
from provero.core.compiler import CheckConfig
from provero.core.results import CheckResult, Severity, Status
from provero.core.sql import quote_identifier


def _parse_duration(duration_str: str) -> int:
    """Parse duration string (e.g., '24h', '30m', '7d') to seconds."""
    match = re.match(r"^(\d+)([smhd])$", duration_str.strip())
    if not match:
        msg = f"Invalid duration format: {duration_str}. Use format like '24h', '30m', '7d'."
        raise ValueError(msg)

    value = int(match.group(1))
    unit = match.group(2)

    multipliers = {"s": 1, "m": 60, "h": 3600, "d": 86400}
    return value * multipliers[unit]


def _format_duration(seconds: int) -> str:
    """Format seconds into human-readable duration."""
    if seconds < 60:
        return f"{seconds}s"
    if seconds < 3600:
        return f"{seconds // 60}m"
    if seconds < 86400:
        return f"{seconds // 3600}h {(seconds % 3600) // 60}m"
    return f"{seconds // 86400}d {(seconds % 86400) // 3600}h"


@register_check("freshness")
def check_freshness(
    connection: Connection,
    table: str,
    check_config: CheckConfig,
) -> CheckResult:
    """Check that data is fresh (most recent row within max_age)."""
    col = check_config.column or ""
    max_age_str = check_config.params.get("max_age", "24h")
    max_age_seconds = _parse_duration(max_age_str)
    qtable = quote_identifier(table)
    qcol = quote_identifier(col)

    # Use epoch() which works on DuckDB, and EXTRACT(EPOCH FROM ...) as fallback
    # for PostgreSQL/other databases.
    try:
        result = connection.execute(
            f"SELECT MAX({qcol}) as latest, "
            f"epoch(CURRENT_TIMESTAMP) - epoch(MAX({qcol})) as age_seconds "
            f"FROM {qtable}"
        )
    except Exception:
        result = connection.execute(
            f"SELECT MAX({qcol}) as latest, "
            f"EXTRACT(EPOCH FROM (CURRENT_TIMESTAMP - MAX({qcol}))) as age_seconds "
            f"FROM {qtable}"
        )

    row = result[0]
    raw_age = row["age_seconds"]
    age_seconds = int(float(raw_age)) if raw_age is not None else None

    if age_seconds is None:
        return CheckResult(
            check_name=f"freshness:{col}",
            check_type="freshness",
            status=Status.FAIL,
            severity=(
                Severity(check_config.severity) if check_config.severity else Severity.CRITICAL
            ),
            column=col,
            observed_value="no data",
            expected_value=f"< {max_age_str}",
        )

    return CheckResult(
        check_name=f"freshness:{col}",
        check_type="freshness",
        status=Status.PASS if age_seconds <= max_age_seconds else Status.FAIL,
        severity=Severity(check_config.severity) if check_config.severity else Severity.CRITICAL,
        column=col,
        observed_value=f"{_format_duration(age_seconds)} ago",
        expected_value=f"< {max_age_str}",
    )


@register_check("latency")
def check_latency(
    connection: Connection,
    table: str,
    check_config: CheckConfig,
) -> CheckResult:
    """Check that the latency between two timestamp columns is within bounds.

    Measures the time difference between a source timestamp (e.g., event_time)
    and a target timestamp (e.g., loaded_at). Useful for detecting pipeline delays.
    """
    source_col: str = check_config.params.get("source_column", check_config.column or "")
    target_col = check_config.params.get("target_column", "")
    max_latency_str = check_config.params.get("max_latency", "1h")
    max_latency_seconds = _parse_duration(max_latency_str)

    severity = Severity(check_config.severity) if check_config.severity else Severity.WARNING

    if not target_col:
        return CheckResult(
            check_name=f"latency:{source_col}",
            check_type="latency",
            status=Status.ERROR,
            severity=severity,
            column=source_col,
            observed_value="target_column parameter is required",
            expected_value=f"< {max_latency_str}",
        )

    qtable = quote_identifier(table)
    qsource = quote_identifier(source_col)
    qtarget = quote_identifier(target_col)

    # Compute average and max latency between the two columns
    try:
        result = connection.execute(
            f"SELECT "
            f"AVG(epoch({qtarget}) - epoch({qsource})) as avg_latency, "
            f"MAX(epoch({qtarget}) - epoch({qsource})) as max_lat, "
            f"COUNT(*) as total "
            f"FROM {qtable} WHERE {qsource} IS NOT NULL AND {qtarget} IS NOT NULL"
        )
    except Exception:
        result = connection.execute(
            f"SELECT "
            f"AVG(EXTRACT(EPOCH FROM ({qtarget} - {qsource}))) as avg_latency, "
            f"MAX(EXTRACT(EPOCH FROM ({qtarget} - {qsource}))) as max_lat, "
            f"COUNT(*) as total "
            f"FROM {qtable} WHERE {qsource} IS NOT NULL AND {qtarget} IS NOT NULL"
        )

    row = result[0]
    avg_latency = float(row["avg_latency"]) if row["avg_latency"] is not None else None
    max_lat = float(row["max_lat"]) if row["max_lat"] is not None else None
    total = row["total"]

    if avg_latency is None or total == 0:
        return CheckResult(
            check_name=f"latency:{source_col}",
            check_type="latency",
            status=Status.FAIL,
            severity=severity,
            column=source_col,
            observed_value="no data",
            expected_value=f"< {max_latency_str}",
        )

    passed = max_lat is not None and max_lat <= max_latency_seconds

    return CheckResult(
        check_name=f"latency:{source_col}",
        check_type="latency",
        status=Status.PASS if passed else Status.FAIL,
        severity=severity,
        column=source_col,
        observed_value=(
            f"avg={_format_duration(int(avg_latency or 0))}, "
            f"max={_format_duration(int(max_lat or 0))}"
        ),
        expected_value=f"< {max_latency_str}",
        row_count=total,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Anomaly detection algorithms using stdlib only."""

from __future__ import annotations

import statistics

from provero.anomaly.models import SENSITIVITY_THRESHOLDS, AnomalyResult

MIN_DATA_POINTS = 5


def detect_zscore(
    values: list[float],
    current: float,
    threshold: float = 3.0,
) -> AnomalyResult:
    """Detect anomalies using Z-Score method.

    Uses mean and standard deviation. Simple but sensitive to outliers
    in the historical data.
    """
    if len(values) < MIN_DATA_POINTS:
        return AnomalyResult(
            is_anomaly=False,
            observed_value=current,
            method="zscore",
            explanation="Insufficient history",
            data_points=len(values),
        )

    mean = statistics.mean(values)
    stdev = statistics.stdev(values)

    if stdev == 0:
        is_anomaly = current != mean
        return AnomalyResult(
            is_anomaly=is_anomaly,
            anomaly_score=1.0 if is_anomaly else 0.0,
            observed_value=current,
            expected_range=(mean, mean),
            method="zscore",
            explanation="All historical values identical; any deviation is anomalous"
            if is_anomaly
            else "Value matches constant history",
            data_points=len(values),
        )

    z = abs(current - mean) / stdev
    score = min(z / threshold, 1.0)
    lower = mean - threshold * stdev
    upper = mean + threshold * stdev

    return AnomalyResult(
        is_anomaly=z > threshold,
        anomaly_score=round(score, 4),
        observed_value=current,
        expected_range=(round(lower, 4), round(upper, 4)),
        method="zscore",
        explanation=f"Z-score={z:.2f}, threshold={threshold}",
        data_points=len(values),
    )


def detect_mad(
    values: list[float],
    current: float,
    threshold: float = 3.0,
) -> AnomalyResult:
    """Detect anomalies using Median Absolute Deviation (MAD).

    More robust to outliers than Z-Score. Default method.
    MAD = median(|xi - median(x)|) * 1.4826
    """
    if len(values) < MIN_DATA_POINTS:
        return AnomalyResult(
            is_anomaly=False,
            observed_value=current,
            method="mad",
            explanation="Insufficient history",
            data_points=len(values),
        )

    med = statistics.median(values)
    deviations = [abs(v - med) for v in values]
    mad = statistics.median(deviations) * 1.4826

    if mad == 0:
        is_anomaly = current != med
        return AnomalyResult(
            is_anomaly=is_anomaly,
            anomaly_score=1.0 if is_anomaly else 0.0,
            observed_value=current,
            expected_range=(med, med),
            method="mad",
            explanation="All historical values identical; any deviation is anomalous"
            if is_anomaly
            else "Value matches constant history",
            data_points=len(values),
        )

    modified_z = abs(current - med) / mad
    score = min(modified_z / threshold, 1.0)
    lower = med - threshold * mad
    upper = med + threshold * mad

    return AnomalyResult(
        is_anomaly=modified_z > threshold,
        anomaly_score=round(score, 4),
        observed_value=current,
        expected_range=(round(lower, 4), round(upper, 4)),
        method="mad",
        explanation=f"Modified Z-score={modified_z:.2f}, threshold={threshold}",
        data_points=len(values),
    )


def detect_iqr(
    values: list[float],
    current: float,
    threshold: float = 1.5,
) -> AnomalyResult:
    """Detect anomalies using Interquartile Range (IQR).

    Uses Q1/Q3 with fence k * IQR. Good for skewed distributions.
    """
    if len(values) < MIN_DATA_POINTS:
        return AnomalyResult(
            is_anomaly=False,
            observed_value=current,
            method="iqr",
            explanation="Insufficient history",
            data_points=len(values),
        )

    sorted_vals = sorted(values)
    n = len(sorted_vals)
    q1_idx = n // 4
    q3_idx = (3 * n) // 4
    q1 = sorted_vals[q1_idx]
    q3 = sorted_vals[q3_idx]
    iqr = q3 - q1

    if iqr == 0:
        is_anomaly = current < q1 or current > q3
        return AnomalyResult(
            is_anomaly=is_anomaly,
            anomaly_score=1.0 if is_anomaly else 0.0,
            observed_value=current,
            expected_range=(q1, q3),
            method="iqr",
            explanation="IQR is zero; values are concentrated"
            if is_anomaly
            else "Value within concentrated range",
            data_points=len(values),
        )

    lower = q1 - threshold * iqr
    upper = q3 + threshold * iqr
    is_anomaly = current < lower or current > upper

    if is_anomaly:
        distance = max(lower - current, current - upper, 0)
        score = min(distance / (threshold * iqr), 1.0)
    else:
        score = 0.0

    return AnomalyResult(
        is_anomaly=is_anomaly,
        anomaly_score=round(score, 4),
        observed_value=current,
        expected_range=(round(lower, 4), round(upper, 4)),
        method="iqr",
        explanation=f"IQR={iqr:.2f}, fence={threshold}",
        data_points=len(values),
    )


_DETECTORS = {
    "zscore": detect_zscore,
    "mad": detect_mad,
    "iqr": detect_iqr,
}


def detect_anomaly(
    values: list[float],
    current: float,
    method: str = "mad",
    sensitivity: str = "medium",
) -> AnomalyResult:
    """Dispatch anomaly detection to the appropriate method.

    Args:
        values: Historical metric values (oldest to newest).
        current: The current value to check.
        method: Detection method ("zscore", "mad", "iqr").
        sensitivity: Sensitivity level ("low", "medium", "high").

    Returns:
        AnomalyResult with detection details.
    """
    detector = _DETECTORS.get(method)
    if detector is None:
        available = ", ".join(sorted(_DETECTORS.keys()))
        return AnomalyResult(
            is_anomaly=False,
            observed_value=current,
            method=method,
            explanation=f"Unknown method '{method}'. Available: {available}",
            data_points=len(values),
        )

    threshold = SENSITIVITY_THRESHOLDS.get(sensitivity, 3.0)
    result = detector(values, current, threshold)
    result.sensitivity = sensitivity
    return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DuckDB connector for local files (Parquet, CSV, JSON)."""

from __future__ import annotations

from typing import Any, cast

import duckdb

from provero.core.sql import is_expression, quote_identifier


class DuckDBConnection:
    """DuckDB connection wrapper."""

    def __init__(self, conn: duckdb.DuckDBPyConnection) -> None:
        self._conn = conn

    def execute(self, query: str, params: dict[str, Any] | None = None) -> list[dict[str, Any]]:
        result = self._conn.execute(query)
        columns = [desc[0] for desc in result.description]
        return [dict(zip(columns, row, strict=True)) for row in result.fetchall()]

    def get_columns(self, table: str) -> list[dict[str, Any]]:
        if is_expression(table):
            result = self._conn.execute(f"DESCRIBE SELECT * FROM {table}")
        else:
            result = self._conn.execute(f"DESCRIBE {quote_identifier(table)}")
        return [
            {"name": row[0], "type": row[1], "nullable": row[2] == "YES"}
            for row in result.fetchall()
        ]


class DuckDBConnector:
    """Connector for DuckDB (local files and in-memory)."""

    def __init__(self, database: str = ":memory:") -> None:
        self.database = database

    def connect(self) -> DuckDBConnection:
        conn = duckdb.connect(self.database)
        return DuckDBConnection(conn)

    def disconnect(self, connection: DuckDBConnection) -> None:
        connection._conn.close()

    def get_schema(self, connection: DuckDBConnection, table: str) -> list[dict[str, Any]]:
        return connection.get_columns(table)

    def get_profile(
        self,
        connection: DuckDBConnection,
        table: str,
        columns: list[str] | None = None,
        sample_size: int | None = None,
    ) -> dict[str, Any]:
        from provero.core.profiler import profile_table

        result = profile_table(connection, table, sample_size=sample_size)
        data = {
            "table": result.table,
            "row_count": result.row_count,
            "column_count": result.column_count,
            "columns": [
                {
                    "name": c.name,
                    "dtype": c.dtype,
                    "null_count": c.null_count,
                    "null_pct": c.null_pct,
                    "distinct_count": c.distinct_count,
                    "distinct_pct": c.distinct_pct,
                }
                for c in result.columns
            ],
        }
        if columns:
            cols_list = cast(list[dict[str, Any]], data["columns"])
            data["columns"] = [c for c in cols_list if c["name"] in columns]
        return data



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Tests for data contracts: validation, diff, and compiler parsing."""

from __future__ import annotations

import textwrap

from provero.contracts.diff import diff_contracts
from provero.contracts.models import (
    ColumnContract,
    ContractConfig,
    SchemaContract,
    SLAConfig,
    ViolationAction,
)
from provero.contracts.validator import validate_contract


class TestSchemaValidation:
    def test_schema_pass(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer"),
                    ColumnContract(name="customer_id", type="varchar"),
                    ColumnContract(name="amount", type="decimal"),
                    ColumnContract(name="status", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert result.status == "pass"
        assert len(result.violations) == 0

    def test_missing_column(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer"),
                    ColumnContract(name="nonexistent_col", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert any(d.change_type == "removed" for d in result.schema_drift)
        assert any("missing" in v.message.lower() for v in result.violations)

    def test_extra_column(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        # Extra columns in table should show as "added" drift
        added = [d for d in result.schema_drift if d.change_type == "added"]
        assert len(added) > 0

    def test_type_mismatch(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="varchar"),  # actual is integer
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert any(d.change_type == "type_changed" for d in result.schema_drift)

    def test_column_checks(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer", checks=["not_null", "unique"]),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        # order_id has no nulls and is unique, so should pass
        assert result.status == "pass"

    def test_parametrized_column_checks(self, duckdb_orders):
        """Column checks with params like range and accepted_values."""
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(
                        name="status",
                        type="varchar",
                        checks=[
                            {
                                "accepted_values": {
                                    "values": ["delivered", "shipped", "pending", "cancelled"],
                                }
                            },
                        ],
                    ),
                    ColumnContract(
                        name="amount",
                        type="decimal",
                        checks=[
                            {"range": {"min": 0, "max": 1000000}},
                        ],
                    ),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert result.status == "pass"

    def test_parametrized_check_fails(self, duckdb_orders):
        """Parametrized check that should fail."""
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(
                        name="status",
                        type="varchar",
                        checks=[
                            {
                                "accepted_values": {
                                    "values": ["delivered", "shipped"],
                                }
                            },  # missing pending, cancelled
                        ],
                    ),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert any("accepted_values" in v.rule for v in result.violations)


class TestSLAValidation:
    def test_availability_pass(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            sla=SLAConfig(availability="true"),
        )
        result = validate_contract(contract, duckdb_orders)
        assert not any(v.rule == "sla.availability" for v in result.violations)

    def test_freshness_with_timestamp(self, duckdb_orders):
        contract = ContractConfig(
            name="events_contract",
            table="events",
            sla=SLAConfig(freshness="24h"),
        )
        result = validate_contract(contract, duckdb_orders)
        # Events have recent timestamps, should pass
        assert not any(v.rule == "sla.freshness" for v in result.violations)

    def test_completeness_pass(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            sla=SLAConfig(completeness="90%"),
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id"),
                    ColumnContract(name="status"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert not any(v.rule == "sla.completeness" for v in result.violations)


class TestOnViolation:
    def test_block_makes_critical(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            on_violation=ViolationAction.BLOCK,
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="nonexistent", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert result.status == "fail"
        assert any(v.severity == "critical" for v in result.violations)

    def test_warn_makes_warning(self, duckdb_orders):
        contract = ContractConfig(
            name="orders_contract",
            table="orders",
            on_violation=ViolationAction.WARN,
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="nonexistent", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert result.status == "warn"
        assert any(v.severity == "warning" for v in result.violations)

    def test_no_table(self):
        """Contract with no table should fail immediately."""
        contract = ContractConfig(name="bad_contract", table="")
        result = validate_contract(contract, None)  # type: ignore[arg-type]
        assert result.status == "fail"


class TestContractDiff:
    def test_no_changes(self):
        contract = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                ]
            ),
        )
        changes = diff_contracts(contract, contract)
        assert len(changes) == 0

    def test_column_added(self):
        old = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                ]
            ),
        )
        new = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                    ColumnContract(name="name", type="varchar"),
                ]
            ),
        )
        changes = diff_contracts(old, new)
        added = [c for c in changes if c.change_type == "added" and "name" in c.field]
        assert len(added) == 1
        assert not added[0].is_breaking

    def test_column_removed_is_breaking(self):
        old = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                    ColumnContract(name="name", type="varchar"),
                ]
            ),
        )
        new = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                ]
            ),
        )
        changes = diff_contracts(old, new)
        removed = [c for c in changes if c.change_type == "removed"]
        assert len(removed) == 1
        assert removed[0].is_breaking

    def test_type_change_is_breaking(self):
        old = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="integer"),
                ]
            ),
        )
        new = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="id", type="varchar"),
                ]
            ),
        )
        changes = diff_contracts(old, new)
        type_changes = [c for c in changes if "type" in c.field]
        assert len(type_changes) == 1
        assert type_changes[0].is_breaking

    def test_sla_change(self):
        old = ContractConfig(name="test", table="orders", sla=SLAConfig(freshness="24h"))
        new = ContractConfig(name="test", table="orders", sla=SLAConfig(freshness="1h"))
        changes = diff_contracts(old, new)
        sla_changes = [c for c in changes if "freshness" in c.field]
        assert len(sla_changes) == 1

    def test_on_violation_change_to_block_is_breaking(self):
        old = ContractConfig(name="test", table="orders", on_violation=ViolationAction.WARN)
        new = ContractConfig(name="test", table="orders", on_violation=ViolationAction.BLOCK)
        changes = diff_contracts(old, new)
        violation_changes = [c for c in changes if c.field == "on_violation"]
        assert len(violation_changes) == 1
        assert violation_changes[0].is_breaking is True

    def test_on_violation_change_to_warn_not_breaking(self):
        old = ContractConfig(name="test", table="orders", on_violation=ViolationAction.BLOCK)
        new = ContractConfig(name="test", table="orders", on_violation=ViolationAction.WARN)
        changes = diff_contracts(old, new)
        violation_changes = [c for c in changes if c.field == "on_violation"]
        assert len(violation_changes) == 1
        assert violation_changes[0].is_breaking is False


class TestCompilerParsing:
    def test_simple_format_with_contracts(self, tmp_path):
        """Contracts should be parsed even in simple format (source + checks at top)."""
        config_path = tmp_path / "provero.yaml"
        config_path.write_text(
            textwrap.dedent("""\
            source:
              type: duckdb
              table: orders

            checks:
              - not_null: order_id

            contracts:
              - name: orders_contract
                table: orders
                schema:
                  columns:
                    - name: order_id
                      type: integer
        """)
        )

        from provero.core.compiler import compile_file

        config = compile_file(config_path)
        assert len(config.contracts) == 1
        assert config.contracts[0].name == "orders_contract"

    def test_parse_contracts_from_yaml(self, tmp_path):
        config_path = tmp_path / "provero.yaml"
        config_path.write_text(
            textwrap.dedent("""\
            version: "1.0"
            suites:
              - name: orders_checks
                source:
                  type: duckdb
                  table: orders
                checks:
                  - not_null: order_id

            contracts:
              - name: orders_contract
                owner: data-team
                table: orders
                on_violation: warn
                schema:
                  columns:
                    - name: order_id
                      type: integer
                      checks: [not_null, unique]
                    - name: status
                      type: varchar
                sla:
                  freshness: 24h
                  completeness: "95%"
        """)
        )

        from provero.core.compiler import compile_file

        config = compile_file(config_path)
        assert len(config.contracts) == 1
        contract = config.contracts[0]
        assert contract.name == "orders_contract"
        assert contract.owner == "data-team"
        assert contract.on_violation == ViolationAction.WARN
        assert len(contract.schema_def.columns) == 2
        assert contract.sla.freshness == "24h"
        assert contract.sla.completeness == "95%"


class TestContractLiveDB:
    def test_full_validation_with_duckdb(self, duckdb_orders):
        """Full contract validation against live DuckDB."""
        contract = ContractConfig(
            name="orders_full",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer"),
                    ColumnContract(name="customer_id", type="varchar"),
                    ColumnContract(name="amount", type="decimal"),
                    ColumnContract(name="status", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        assert result.status == "pass"

    def test_schema_drift_extra_column(self, duckdb_orders):
        """ALTER TABLE adds a column, contract should detect drift."""
        duckdb_orders._conn.execute("ALTER TABLE orders ADD COLUMN region VARCHAR")
        contract = ContractConfig(
            name="orders_drift",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(name="order_id", type="integer"),
                    ColumnContract(name="customer_id", type="varchar"),
                    ColumnContract(name="amount", type="decimal"),
                    ColumnContract(name="status", type="varchar"),
                ]
            ),
        )
        result = validate_contract(contract, duckdb_orders)
        added = [d for d in result.schema_drift if d.change_type == "added"]
        assert any("region" in d.column for d in added)

    def test_sla_freshness_live(self, duckdb_orders):
        """Events with recent timestamps should pass freshness SLA."""
        contract = ContractConfig(
            name="events_fresh",
            table="events",
            sla=SLAConfig(freshness="24h"),
        )
        result = validate_contract(contract, duckdb_orders)
        assert not any(v.rule == "sla.freshness" for v in result.violations)

    def test_diff_with_dict_checks(self):
        """Contract diff handles parametrized (dict) checks in columns."""
        old = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(
                        name="status",
                        type="varchar",
                        checks=[{"accepted_values": {"values": ["a", "b"]}}],
                    ),
                ]
            ),
        )
        new = ContractConfig(
            name="test",
            table="orders",
            schema_def=SchemaContract(
                columns=[
                    ColumnContract(
                        name="status",
                        type="varchar",
                        checks=[{"accepted_values": {"values": ["a", "b", "c"]}}],
                    ),
                ]
            ),
        )
        changes = diff_contracts(old, new)
        # Should detect the check change without crashing
        assert isinstance(changes, list)

    def test_multiple_contracts_parsed(self, tmp_path):
        """YAML with 2 contracts, compile_file parses both."""
        config_path = tmp_path / "provero.yaml"
        config_path.write_text(
            textwrap.dedent("""\
            source:
              type: duckdb
              table: orders

            checks:
              - not_null: order_id

            contracts:
              - name: contract_one
                table: orders
                schema:
                  columns:
                    - name: order_id
                      type: integer
              - name: contract_two
                table: events
                schema:
                  columns:
                    - name: event_id
                      type: integer
        """)
        )

        from provero.core.compiler import compile_file

        config = compile_file(config_path)
        assert len(config.contracts) == 2
        names = {c.name for c in config.contracts}
        assert names == {"contract_one", "contract_two"}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Tests for the check execution engine."""

from __future__ import annotations

import pytest

from provero.connectors.duckdb import DuckDBConnection, DuckDBConnector
from provero.core.compiler import CheckConfig, SourceConfig, SuiteConfig
from provero.core.engine import run_suite
from provero.core.results import Severity, Status


class _SharedDuckDBConnector:
    """A connector that always returns the same connection (for testing)."""

    def __init__(self, conn: DuckDBConnection) -> None:
        self._conn = conn

    def connect(self) -> DuckDBConnection:
        return self._conn

    def disconnect(self, connection: DuckDBConnection) -> None:
        pass  # Don't close, fixture manages lifecycle


@pytest.fixture
def orders_connector():
    connector = DuckDBConnector()
    conn = connector.connect()
    conn._conn.execute("""
        CREATE TABLE orders (
            order_id INTEGER,
            customer_id VARCHAR,
            amount DECIMAL(10,2),
            status VARCHAR
        )
    """)
    conn._conn.execute("""
        INSERT INTO orders VALUES
        (1, 'C001', 150.00, 'delivered'),
        (2, 'C002', 89.99, 'shipped'),
        (3, 'C003', 220.50, 'pending'),
        (4, 'C001', 45.00, 'delivered'),
        (5, 'C004', 999.99, 'cancelled')
    """)
    yield _SharedDuckDBConnector(conn)
    connector.disconnect(conn)


def _make_suite(checks: list[CheckConfig], table: str = "orders") -> SuiteConfig:
    return SuiteConfig(
        name="test_suite",
        source=SourceConfig(type="duckdb", table=table),
        checks=checks,
    )


class TestRunSuiteOptimized:
    def test_all_pass(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", columns=["order_id", "customer_id"]),
                CheckConfig(check_type="unique", column="order_id"),
                CheckConfig(check_type="row_count", params={"min": 1, "max": 100}),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)

        assert result.status == Status.PASS
        assert result.total >= 3
        assert result.passed == result.total
        assert result.failed == 0
        assert result.quality_score == 100.0

    def test_mixed_pass_fail(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id"),
                CheckConfig(check_type="unique", column="customer_id"),  # C001 duplicated
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)

        assert result.status == Status.FAIL
        assert result.passed >= 1
        assert result.failed >= 1
        assert 0 < result.quality_score < 100

    def test_run_id_consistent_across_checks(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id"),
                CheckConfig(check_type="row_count", params={"min": 1}),
            ]
        )
        result = run_suite(suite, orders_connector)
        run_ids = {c.run_id for c in result.checks}
        assert len(run_ids) == 1
        assert "" not in run_ids

    def test_duration_tracked(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="row_count", params={"min": 1}),
            ]
        )
        result = run_suite(suite, orders_connector)
        assert result.duration_ms >= 0

    def test_severity_propagated_through_optimizer(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id", severity="warning"),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)
        assert result.checks[0].severity == Severity.WARNING


class TestRunSuiteUnoptimized:
    def test_all_pass_no_optimize(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id"),
                CheckConfig(check_type="unique", column="order_id"),
                CheckConfig(check_type="row_count", params={"min": 1}),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=False)

        assert result.status == Status.PASS
        assert result.total == 3
        assert result.passed == 3

    def test_non_batchable_checks(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(
                    check_type="custom_sql",
                    params={
                        "name": "positive_amounts",
                        "query": "SELECT COUNT(*) = 0 FROM orders WHERE amount < 0",
                    },
                ),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)
        assert result.total == 1
        assert result.checks[0].check_type == "custom_sql"
        assert result.checks[0].status == Status.PASS


class TestRunSuiteErrorHandling:
    def test_unknown_check_type_lists_available(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="nonexistent_check", column="order_id"),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=False)

        assert result.total == 1
        assert result.checks[0].status == Status.ERROR
        assert "nonexistent_check" in result.checks[0].observed_value
        assert "Available types:" in result.checks[0].observed_value

    def test_missing_table_error(self, orders_connector):
        suite = _make_suite(
            [CheckConfig(check_type="row_count", params={"min": 1})],
            table="table_that_does_not_exist",
        )
        result = run_suite(suite, orders_connector, optimize=True)

        assert result.status == Status.FAIL or result.status == Status.PASS
        # Should have an error in checks
        error_checks = [c for c in result.checks if c.status == Status.ERROR]
        assert len(error_checks) >= 1

    def test_error_does_not_crash_other_checks(self, orders_connector):
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id"),
                CheckConfig(
                    check_type="custom_sql",
                    params={"name": "bad_query", "query": "THIS IS NOT SQL"},
                ),
                CheckConfig(check_type="row_count", params={"min": 1}),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)

        # Should have at least the batchable checks + the error
        assert result.total >= 2
        statuses = [c.status for c in result.checks]
        assert Status.PASS in statuses
        assert Status.ERROR in statuses


class TestRunSuiteMixedBatchAndIndividual:
    def test_batch_plus_individual(self, orders_connector):
        """Batchable and non-batchable checks both execute correctly."""
        suite = _make_suite(
            [
                CheckConfig(check_type="not_null", column="order_id"),
                CheckConfig(check_type="unique", column="order_id"),
                CheckConfig(check_type="row_count", params={"min": 1}),
                CheckConfig(
                    check_type="custom_sql",
                    params={
                        "name": "no_neg",
                        "query": "SELECT COUNT(*) = 0 FROM orders WHERE amount < 0",
                    },
                ),
            ]
        )
        result = run_suite(suite, orders_connector, optimize=True)

        check_types = {c.check_type for c in result.checks}
        assert "not_null" in check_types
        assert "unique" in check_types
        assert "row_count" in check_types
        assert "custom_sql" in check_types
        assert result.status == Status.PASS



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Webhook alert sender."""

from __future__ import annotations

import json
import os
import urllib.error
import urllib.request
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from provero.alerts.models import AlertConfig
    from provero.core.results import SuiteResult


def _resolve_env_vars(value: str) -> str:
    """Expand ${ENV_VAR} references in a string."""
    if value.startswith("${") and value.endswith("}"):
        env_var = value[2:-1]
        resolved = os.environ.get(env_var)
        if resolved is None:
            msg = f"Environment variable {env_var} is not set"
            raise ValueError(msg)
        return resolved
    return os.path.expandvars(value)


def _should_fire(alert: AlertConfig, result: SuiteResult) -> bool:
    """Determine if an alert should fire based on the trigger condition."""
    from provero.core.results import Status

    trigger = alert.trigger.lower()
    if trigger == "on_failure":
        return result.status == Status.FAIL
    if trigger == "always":
        return True
    if trigger == "on_success":
        return result.status == Status.PASS
    return result.status == Status.FAIL


def _build_payload(result: SuiteResult) -> dict:
    """Build the JSON payload for a webhook alert."""
    failed_checks = [
        {
            "check": c.check_name,
            "type": c.check_type,
            "column": c.column,
            "observed": str(c.observed_value),
            "expected": str(c.expected_value),
        }
        for c in result.checks
        if c.status.value in ("fail", "error")
    ]

    return {
        "suite": result.suite_name,
        "status": result.status.value,
        "quality_score": result.quality_score,
        "total": result.total,
        "passed": result.passed,
        "failed": result.failed,
        "errored": result.errored,
        "duration_ms": result.duration_ms,
        "timestamp": result.started_at.isoformat(),
        "failed_checks": failed_checks,
    }


def send_alert(alert: AlertConfig, result: SuiteResult) -> bool:
    """Send a single webhook alert. Returns True on success."""
    if not _should_fire(alert, result):
        return False

    url = _resolve_env_vars(alert.url)
    headers = {k: _resolve_env_vars(v) for k, v in alert.headers.items()}
    headers.setdefault("Content-Type", "application/json")

    payload = json.dumps(_build_payload(result)).encode("utf-8")
    req = urllib.request.Request(url, data=payload, headers=headers, method="POST")

    try:
        with urllib.request.urlopen(req, timeout=10) as resp:
            return 200 <= resp.status < 300
    except (urllib.error.URLError, urllib.error.HTTPError, OSError):
        return False


def send_alerts(
    alerts: list[AlertConfig],
    result: SuiteResult,
) -> list[bool]:
    """Send all configured alerts for a suite result.

    Returns a list of booleans indicating success/failure for each alert.
    """
    return [send_alert(alert, result) for alert in alerts]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Tests for graceful degradation when optional dependencies are missing."""

from __future__ import annotations

import importlib
import inspect
import sys
from unittest.mock import patch


class TestGracefulDegradation:
    def test_core_import_without_pandas(self):
        """provero.core should import even without pandas installed."""
        with patch.dict(sys.modules, {"pandas": None}):
            # Force re-import
            mod = importlib.import_module("provero.core.compiler")
            assert hasattr(mod, "compile_file")

    def test_core_import_without_scipy(self):
        """provero.anomaly should work without scipy (uses stdlib statistics)."""
        with patch.dict(sys.modules, {"scipy": None, "scipy.stats": None}):
            mod = importlib.import_module("provero.anomaly.detectors")
            assert hasattr(mod, "detect_zscore")

    def test_anomaly_detectors_use_stdlib_only(self):
        """Anomaly detectors should not import scipy or numpy."""
        from provero.anomaly import detectors

        source = inspect.getsource(detectors)
        # Detectors should use stdlib statistics, not scipy/numpy
        assert "import scipy" not in source
        assert "import numpy" not in source
        assert "from scipy" not in source
        assert "from numpy" not in source

    def test_dataframe_connector_needs_duckdb(self):
        """DataFrameConnector uses duckdb internally and should work."""
        # duckdb is a required dep, not optional

        from provero.connectors.dataframe import DataFrameConnector

        assert DataFrameConnector is not None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Unit tests for anomaly detection algorithms."""

from __future__ import annotations

from provero.anomaly.detectors import detect_anomaly, detect_iqr, detect_mad, detect_zscore


class TestZScore:
    def test_normal_value(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_zscore(values, 100.0, threshold=3.0)
        assert not result.is_anomaly
        assert result.method == "zscore"
        assert result.data_points == 8

    def test_anomalous_value(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_zscore(values, 200.0, threshold=3.0)
        assert result.is_anomaly
        assert result.anomaly_score > 0

    def test_insufficient_data(self):
        values = [1.0, 2.0, 3.0]
        result = detect_zscore(values, 100.0)
        assert not result.is_anomaly
        assert "Insufficient" in result.explanation

    def test_constant_values_same(self):
        values = [5.0] * 10
        result = detect_zscore(values, 5.0)
        assert not result.is_anomaly

    def test_constant_values_different(self):
        values = [5.0] * 10
        result = detect_zscore(values, 6.0)
        assert result.is_anomaly
        assert result.anomaly_score == 1.0


class TestMAD:
    def test_normal_value(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_mad(values, 100.0, threshold=3.0)
        assert not result.is_anomaly
        assert result.method == "mad"

    def test_anomalous_value(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_mad(values, 200.0, threshold=3.0)
        assert result.is_anomaly

    def test_robust_to_outliers(self):
        """MAD should handle datasets with existing outliers better than Z-Score."""
        values = [10.0, 11.0, 10.5, 10.2, 9.8, 100.0, 10.3, 10.1]  # 100 is an outlier
        result = detect_mad(values, 10.0)
        assert not result.is_anomaly

    def test_insufficient_data(self):
        result = detect_mad([1.0, 2.0], 5.0)
        assert not result.is_anomaly

    def test_constant_values(self):
        values = [42.0] * 10
        result = detect_mad(values, 43.0)
        assert result.is_anomaly


class TestIQR:
    def test_normal_value(self):
        values = list(range(20))
        result = detect_iqr([float(v) for v in values], 10.0)
        assert not result.is_anomaly
        assert result.method == "iqr"

    def test_anomalous_value(self):
        values = [float(v) for v in range(20)]
        result = detect_iqr(values, 100.0, threshold=1.5)
        assert result.is_anomaly

    def test_insufficient_data(self):
        result = detect_iqr([1.0, 2.0], 50.0)
        assert not result.is_anomaly

    def test_concentrated_values(self):
        values = [5.0] * 10
        result = detect_iqr(values, 5.0)
        assert not result.is_anomaly


class TestDispatcher:
    def test_default_method_is_mad(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_anomaly(values, 100.0)
        assert result.method == "mad"

    def test_sensitivity_low(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_anomaly(values, 108.0, sensitivity="low")
        assert result.sensitivity == "low"
        # Low sensitivity = higher threshold = less likely to flag

    def test_sensitivity_high(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_anomaly(values, 108.0, sensitivity="high")
        assert result.sensitivity == "high"

    def test_unknown_method(self):
        result = detect_anomaly([1.0, 2.0, 3.0, 4.0, 5.0], 3.0, method="invalid")
        assert not result.is_anomaly
        assert "Unknown" in result.explanation

    def test_zscore_via_dispatcher(self):
        values = [100.0, 102.0, 98.0, 101.0, 99.0, 100.5, 101.5, 99.5]
        result = detect_anomaly(values, 100.0, method="zscore")
        assert result.method == "zscore"

    def test_iqr_via_dispatcher(self):
        values = [float(v) for v in range(20)]
        result = detect_anomaly(values, 10.0, method="iqr")
        assert result.method == "iqr"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Tests for JSON Schema validation of provero.yaml."""

from __future__ import annotations

import json
from pathlib import Path

import pytest
import yaml
from jsonschema import ValidationError, validate


@pytest.fixture
def schema():
    # Try workspace root first, then provero-core local
    schema_path = Path(__file__).parent.parent.parent / "aql-spec" / "schema.json"
    if not schema_path.exists():
        schema_path = Path(__file__).parent.parent / "aql-spec" / "schema.json"
    with schema_path.open() as f:
        return json.load(f)


class TestSchemaSimpleFormat:
    def test_minimal_valid(self, schema):
        config = {
            "source": {"type": "duckdb", "table": "orders"},
            "checks": [{"not_null": "id"}],
        }
        validate(instance=config, schema=schema)

    def test_full_simple_format(self, schema):
        config = {
            "version": "1.0",
            "source": {
                "type": "postgres",
                "connection": "postgresql://localhost/db",
                "table": "orders",
            },
            "checks": [
                {"not_null": ["id", "name"]},
                {"unique": "id"},
                {"row_count": {"min": 1, "max": 1000}},
                {"range": {"column": "amount", "min": 0, "max": 100000}},
                {"accepted_values": {"column": "status", "values": ["a", "b"]}},
                {"regex": {"column": "email", "pattern": "^.+@.+$"}},
                {"freshness": {"column": "created_at", "max_age": "24h"}},
                {"completeness": {"column": "email", "min": 0.95}},
                {"custom_sql": "SELECT COUNT(*) = 0 FROM orders WHERE amount < 0"},
                {"type": {"column": "id", "expected": "integer"}},
                {
                    "latency": {
                        "source_column": "event_time",
                        "target_column": "loaded_at",
                        "max_latency": "1h",
                    }
                },
            ],
        }
        validate(instance=config, schema=schema)

    def test_missing_source_fails(self, schema):
        config = {"checks": [{"not_null": "id"}]}
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)

    def test_missing_checks_fails(self, schema):
        config = {"source": {"type": "duckdb"}}
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)

    def test_invalid_source_type_fails(self, schema):
        config = {
            "source": {"type": "oracle"},
            "checks": [{"not_null": "id"}],
        }
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)


class TestSchemaSuiteFormat:
    def test_valid_suite_format(self, schema):
        config = {
            "suites": [
                {
                    "name": "orders_daily",
                    "source": {"type": "duckdb", "table": "orders"},
                    "checks": [{"not_null": "id"}],
                }
            ]
        }
        validate(instance=config, schema=schema)

    def test_suite_with_named_sources(self, schema):
        config = {
            "sources": {
                "warehouse": {"type": "postgres", "connection": "postgresql://localhost/db"},
            },
            "suites": [
                {
                    "name": "orders_daily",
                    "source": "warehouse",
                    "table": "orders",
                    "tags": ["critical"],
                    "checks": [{"unique": "id"}],
                }
            ],
        }
        validate(instance=config, schema=schema)

    def test_suite_missing_name_fails(self, schema):
        config = {
            "suites": [
                {
                    "source": {"type": "duckdb"},
                    "checks": [{"not_null": "id"}],
                }
            ]
        }
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)


class TestSchemaSeverity:
    def test_severity_on_dict_check(self, schema):
        config = {
            "source": {"type": "duckdb", "table": "orders"},
            "checks": [
                {"range": {"column": "amount", "min": 0, "severity": "warning"}},
            ],
        }
        validate(instance=config, schema=schema)

    def test_invalid_severity_fails(self, schema):
        config = {
            "source": {"type": "duckdb", "table": "orders"},
            "checks": [
                {"range": {"column": "amount", "min": 0, "severity": "mega_critical"}},
            ],
        }
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)


class TestSchemaContracts:
    def test_valid_contract(self, schema):
        config = {
            "suites": [
                {
                    "name": "s",
                    "source": {"type": "duckdb"},
                    "checks": [{"not_null": "id"}],
                }
            ],
            "contracts": [
                {
                    "name": "my_contract",
                    "source": {"type": "postgres", "connection": "postgresql://localhost/db"},
                    "owner": "data-team",
                    "on_violation": "block",
                }
            ],
        }
        validate(instance=config, schema=schema)

    def test_invalid_on_violation_fails(self, schema):
        config = {
            "suites": [
                {
                    "name": "s",
                    "source": {"type": "duckdb"},
                    "checks": [{"not_null": "id"}],
                }
            ],
            "contracts": [
                {
                    "name": "c",
                    "source": {"type": "duckdb"},
                    "on_violation": "explode",
                }
            ],
        }
        with pytest.raises(ValidationError):
            validate(instance=config, schema=schema)


class TestSchemaExampleFile:
    def test_quickstart_example_validates(self, schema):
        example_path = (
            Path(__file__).parent.parent.parent / "examples" / "quickstart" / "provero.yaml"
        )
        if not example_path.exists():
            example_path = Path(__file__).parent.parent / "examples" / "quickstart" / "provero.yaml"
        if not example_path.exists():
            pytest.skip("quickstart example not found")
        with example_path.open() as f:
            config = yaml.safe_load(f)
        validate(instance=config, schema=schema)
